@@ -28,12 +28,14 @@
 #include <stdio.h>
 #include <stdarg.h>
 #include <string.h>
-#include <wait.h>
+#include <sys/wait.h>
 #include <unistd.h>
 #ifdef HAVE_SYS_STAT_H
 #include <sys/stat.h>
 #endif
 #include <errno.h>
+#include <assert.h>
+
 #define obstack_chunk_alloc malloc
 #define obstack_chunk_free free
 #include <obstack.h>
@@ -77,6 +79,8 @@
 
 #define DIR_UP ".."
 
+static bool verbose = false;
+
 static const char *outname = NULL;
 
 static void __attribute__ ((format (printf, 1, 2)))
@@ -151,8 +155,6 @@ symbol_hash_lookup (const char *string)
   return (symbol *) *e;
 }
 
-#define COMMENT_PREFIX "#"
-
 typedef enum Kind
 {
   /* 0-ff used for single char tokens */
@@ -177,6 +179,9 @@ typedef struct Token
   char const *ptr;
 } Token;
 
+/* Token of the preamble '.target' directive's argument.  */
+static Token *tok_preamble_target_arg;
+
 /* statement info */
 typedef enum Vis
 {
@@ -864,26 +869,74 @@ traverse (void **slot, void *data)
 }
 
 static void
-process (FILE *in, FILE *out, int verify, const char *outname)
+process (FILE *in, FILE *out, int *verify, const char *inname)
 {
-  symbol_table = htab_create (500, hash_string_hash, hash_string_eq,
-                              NULL);
-
   const char *input = read_file (in);
+
+  /* As expected by GCC, handle an empty input file specially.  See
+     <https://github.com/MentorEmbedded/nvptx-tools/pull/26> "[nvptx-as] Allow
+     empty input file" for reference.  */
+  if (*input == '\0')
+    {
+      /* Produce an empty output file.  */
+
+      /* An empty file isn't a valid PTX file.  */
+      *verify = 0;
+
+      return;
+    }
+
   Token *tok = tokenize (input);
 
-  /* By default, when ptxas is not in PATH, do minimalistic verification,
-     just require that the first non-comment directive is .version.  */
-  if (verify < 0)
+  /* Do minimalistic verification, so that we reliably reject (certain classes
+     of) invalid input.  (If available and applicable, 'ptxas' is later used to
+     verify the whole output file.)  */
+  if (*verify != 0)
     {
-      size_t i;
-      for (i = 0; tok[i].kind == K_comment; i++)
-	;
-      if (tok[i].kind != K_dotted || !is_keyword (&tok[i], "version"))
+      /* Verify the preamble as generated by GCC.  */
+      size_t i = 0;
+      while (tok[i].kind == K_comment)
+	i++;
+      if (tok[i].kind == K_dotted && is_keyword (&tok[i], "version"))
+	i++;
+      else
 	fatal_error ("missing .version directive at start of file '%s'",
-		     outname);
+		     inname);
+      /* 'ptxas' doesn't seem to allow comments or line breaks here, but it's
+	 not documented.  */
+      while (tok[i].kind == K_comment)
+	i++;
+      if (tok[i].kind == K_number)
+	i++;
+      else
+	fatal_error ("malformed .version directive at start of file '%s'",
+		     inname);
+      while (tok[i].kind == K_comment)
+	i++;
+      if (tok[i].kind == K_dotted && is_keyword (&tok[i], "target"))
+	i++;
+      else
+	fatal_error ("missing .target directive at start of file '%s'",
+		     inname);
+      while (tok[i].kind == K_comment)
+	i++;
+      if (tok[i].kind == K_symbol)
+	{
+	  tok_preamble_target_arg = &tok[i];
+	  i++;
+	}
+      else
+	fatal_error ("malformed .target directive at start of file '%s'",
+		     inname);
+      /* PTX allows here a "comma separated list of target specifiers", but GCC
+	 doesn't generate that, and we don't support that.  */
+      if (tok[i].kind == ',')
+	fatal_error ("unsupported list in .target directive at start of file '%s'",
+		     inname);
     }
 
+  symbol_table = htab_create (500, hash_string_hash, hash_string_eq, NULL);
+
   do
     tok = parse_file (tok);
   while (tok->kind);
@@ -935,6 +988,18 @@ do_wait (const char *prog, struct pex_obj *pex)
 static void
 fork_execute (const char *prog, char *const *argv)
 {
+  if (verbose)
+    {
+      for (char *const *arg = argv; *arg; ++arg)
+	{
+	  if (**arg == '\0')
+	    fprintf (stderr, " ''");
+	  else
+	    fprintf (stderr, " %s", *arg);
+	}
+      fprintf (stderr, "\n");
+    }
+
   struct pex_obj *pex = pex_init (0, "nvptx-as", NULL);
   if (pex == NULL)
     fatal_error ("pex_init failed: %m");
@@ -1044,10 +1109,10 @@ int
 main (int argc, char **argv)
 {
   FILE *in = stdin;
+  const char *inname = "{standard input}";
   FILE *out = stdout;
-  bool verbose __attribute__((unused)) = false;
   int verify = -1;
-  const char *smver = "sm_30";
+  const char *target_arg_force = NULL;
 
   int o;
   int option_index = 0;
@@ -1073,7 +1138,7 @@ main (int argc, char **argv)
 	  outname = optarg;
 	  break;
 	case 'm':
-	  smver = optarg;
+	  target_arg_force = optarg;
 	  break;
 	case 'I':
 	  /* Ignore include paths.  */
@@ -1082,6 +1147,8 @@ main (int argc, char **argv)
 	  printf ("\
 Usage: nvptx-none-as [option...] [asmfile]\n\
 Options:\n\
+  -m TARGET             Override target architecture used for ptxas\n\
+                        verification (default: deduce from input's preamble)\n\
   -o FILE               Write output to FILE\n\
   -v                    Be verbose\n\
   --verify              Do verify output is acceptable to ptxas\n\
@@ -1094,7 +1161,7 @@ Report bugs to %s.\n",
 	  exit (0);
 	case 'V':
 	  printf ("\
-nvtpx-none-as %s%s\n\
+nvptx-none-as %s%s\n\
 Copyright %s Mentor Graphics\n\
 License GPLv3+: GNU GPL version 3 or later <http://gnu.org/licenses/gpl.html>\n\
 This program is free software; you may redistribute it under the terms of\n\
@@ -1116,22 +1183,71 @@ This program has absolutely no warranty.\n",
     fatal_error ("cannot open '%s'", outname);
 
   if (argc > optind)
-    in = fopen (argv[optind], "r");
+    {
+      inname = argv[optind];
+      in = fopen (inname, "r");
+    }
   if (!in)
     fatal_error ("cannot open input ptx file");
 
+  process (in, out, &verify, inname);
+
+  if (outname)
+    fclose (out);
+
   if (outname == NULL)
+    /* We don't have a PTX file for 'ptxas' to read in; skip verification.  */
     verify = 0;
   else if (verify == -1)
     if (program_available ("ptxas"))
       verify = 1;
 
-  process (in, out, verify, outname);
-  if (outname)
-    fclose (out);
-
   if (verify > 0)
     {
+      const char *target_arg;
+      if (target_arg_force)
+	target_arg = target_arg_force;
+      else
+	{
+	  assert (tok_preamble_target_arg);
+
+	  /* Override the default '--gpu-name' of 'ptxas': its default may not
+	     be sufficient for what is requested in the '.target' directive in
+	     the input's preamble:
+
+	         ptxas fatal   : SM version specified by .target is higher than default SM version assumed
+
+	     In this case, use the '.target' we found in the preamble.  */
+	  target_arg = xstrndup (tok_preamble_target_arg->ptr,
+				 tok_preamble_target_arg->len);
+
+	  if ((strcmp ("sm_30", target_arg) == 0)
+	      || (strcmp ("sm_32", target_arg) == 0))
+	    {
+	      /* Starting with CUDA 11.0, "Support for Kepler 'sm_30' and
+		 'sm_32' architecture based products is dropped", and these may
+		 no longer be specified in '--gpu-name' of 'ptxas':
+
+		     ptxas fatal   : Value 'sm_30' is not defined for option 'gpu-name'
+
+		     ptxas fatal   : Value 'sm_32' is not defined for option 'gpu-name'
+
+		 ..., but we need to continue supporting GCC emitting
+		 '.target sm_30' code, for example.
+
+		 Detecting the CUDA/'ptxas' version and the supported
+		 '--gpu-name' options is clumsy, so in this case, just use
+		 'sm_35', which is the baseline supported by all current CUDA
+		 versions down to CUDA 6.5, at least.  */
+	      if (verbose)
+		fprintf (stderr, "Verifying %s code", target_arg);
+	      free ((void *) target_arg);
+	      target_arg = "sm_35";
+	      if (verbose)
+		fprintf (stderr, " with %s code generation.\n", target_arg);
+	    }
+	}
+
       struct obstack argv_obstack;
       obstack_init (&argv_obstack);
       obstack_ptr_grow (&argv_obstack, "ptxas");
@@ -1140,11 +1256,17 @@ This program has absolutely no warranty.\n",
       obstack_ptr_grow (&argv_obstack, "/dev/null");
       obstack_ptr_grow (&argv_obstack, outname);
       obstack_ptr_grow (&argv_obstack, "--gpu-name");
-      obstack_ptr_grow (&argv_obstack, smver);
+      obstack_ptr_grow (&argv_obstack, target_arg);
       obstack_ptr_grow (&argv_obstack, "-O0");
       obstack_ptr_grow (&argv_obstack, NULL);
       char *const *new_argv = XOBFINISH (&argv_obstack, char *const *);
       fork_execute (new_argv[0], new_argv);
     }
+  else if (verify < 0)
+    {
+      if (verbose)
+	fprintf (stderr, "'ptxas' not available.\n");
+    }
+
   return 0;
 }
